// This file is part of PUMA.
// Copyright (C) 1999-2003  The PUMA developer team.
//                                                                
// This program is free software;  you can redistribute it and/or 
// modify it under the terms of the GNU General Public License as 
// published by the Free Software Foundation; either version 2 of 
// the License, or (at your option) any later version.            
//                                                                
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
// GNU General Public License for more details.                   
//                                                                
// You should have received a copy of the GNU General Public      
// License along with this program; if not, write to the Free     
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
// MA  02111-1307  USA                                            

#ifndef __CStructure_h__
#define __CStructure_h__

// base class of everything that also is a scope
// (function, class, union, namespace...)
// stores semantic info objects for variables, 
// enumerators, function and template parameters,
// typedefs, classes, unions, enumerations, 
// namespaces, using-directives, friend function 
// declarations, functions, class methods, class 
// data members, function and class templates,
// function and class template instances, member 
// aliases

#include "Puma/CScopeInfo.h"
#include "Puma/Array.h"

#ifndef __puma

#include <map>
#include <set>
#include <list>
#include <string>
#include "Puma/DString.h"
using namespace std;

#endif

namespace Puma {


class CAttributeInfo;
class CTemplateParamInfo;
class CFunctionInfo;
class CClassInfo;
class CUnionInfo;
class CEnumInfo;
class CTypedefInfo;
class CNamespaceInfo;
class CUsingInfo;
class CMemberAliasInfo;

class CStructure : public CScopeInfo {
public:
#ifndef __puma
  typedef list<CObjectInfo*> ObjectInfoList;
  typedef map<DString,ObjectInfoList> ObjectsByName;
  typedef set<CObjectInfo*> ObjectInfoSet;
#endif 

private:
  CStructure                *_Shared;
  Array<CObjectInfo*>        _Objects; // ALL objects (mixed, in the order added to)
  Array<CAttributeInfo*>     _Attributes;
  Array<CTemplateParamInfo*> _TemplateParams;
  Array<CFunctionInfo*>      _Functions;
  Array<CObjectInfo*>        _Types;
  Array<CUsingInfo*>         _Usings;     
  Array<CRecord*>            _Friends;
  Array<CNamespaceInfo*>     _Namespaces;     
  Array<CMemberAliasInfo*>   _Aliases;     
#ifndef __puma
  ObjectsByName              _ObjectMap;
  ObjectInfoSet              _ObjectSet;
#endif

protected:
  CStructure (ObjectId);

public:
  ~CStructure ();

  // Get ... 
#ifndef __puma
  ObjectsByName& ObjectInfoMap() { return _Shared->_ObjectMap; }
  const ObjectInfoSet& ObjectInfos() const { return _Shared->_ObjectSet; }
#endif 
  unsigned Objects () const;                        // all entities in scope
  unsigned Objects (const DString&) const;
  unsigned Types () const;                          // type declarations
  unsigned Types (const DString&) const;
  unsigned Attributes () const;                     // data objects
  unsigned Attributes (const DString&) const;
  unsigned TemplateParams () const;                 // template parameters
  unsigned TemplateParams (const DString&) const;
  unsigned Functions () const;                      // functions, methods, function
  unsigned Functions (const DString&) const;        // templates and its instances
  unsigned Usings () const;                         // using-directives
  unsigned Usings (const DString&) const;
  unsigned Namespaces () const;                     // namespaces
  unsigned Namespaces (const DString&) const;
  unsigned Friends () const;                        // friend function declarations
  unsigned Friends (const DString&) const;
  CObjectInfo *Object (unsigned) const;
  CObjectInfo *Object (const DString&, unsigned = 0) const;
  CObjectInfo *Type (unsigned) const;
  CObjectInfo *Type (const DString&, unsigned = 0) const;
  CUsingInfo *Using (unsigned) const;
  CUsingInfo *Using (const DString&, unsigned = 0) const;
  CNamespaceInfo *Namespace (unsigned) const;
  CNamespaceInfo *Namespace (const DString&, unsigned = 0) const;
  CRecord *Friend (unsigned) const;
  CRecord *Friend (const DString&, unsigned = 0) const;
  CAttributeInfo *Attribute (unsigned) const;
  CAttributeInfo *Attribute (const DString&, unsigned = 0) const;
  CTemplateParamInfo *TemplateParam (unsigned) const;
  CTemplateParamInfo *TemplateParam (const DString&, unsigned = 0) const;
  CFunctionInfo *Function (unsigned) const;
  CFunctionInfo *Function (const DString&, unsigned = 0) const;
  CFunctionInfo *Function (const DString&, CTypeInfo *) const;

  bool isMemberAlias (const CObjectInfo *) const; // is a member alias info?
  CMemberAliasInfo *MemberAlias (const CObjectInfo *) const; // get denoted member

  // Set ...
  void setShared (CStructure *);
  void addObject (CObjectInfo *);
  void addAttribute (CAttributeInfo *);
  void addTemplateParam (CTemplateParamInfo *);
  void addFunction (CFunctionInfo *);
  void addType (CObjectInfo *);
  void addUsing (CUsingInfo *);
  void addNamespace (CNamespaceInfo *);
  void addFriend (CRecord *);
  void removeObject (const CObjectInfo *);       // WILL NOT BE DESTROYED!
  void removeAttribute (const CAttributeInfo *); // WILL NOT BE DESTROYED!
  void removeTemplateParam (const CTemplateParamInfo *); // WILL NOT BE DESTROYED!
  void removeFunction (const CFunctionInfo *);   // WILL NOT BE DESTROYED!
  void removeType (const CObjectInfo *);         // WILL NOT BE DESTROYED!
  void removeUsing (const CUsingInfo *);         // WILL NOT BE DESTROYED!
  void removeNamespace (const CNamespaceInfo *); // WILL NOT BE DESTROYED!
  void removeFriend (const CRecord *);           // WILL NOT BE DESTROYED!
  void removeMemberAlias (const CMemberAliasInfo *); // WILL NOT BE DESTROYED!

  void removeRegisterEntry ();
  
  void addObjectName(CObjectInfo *); // internal use only, do not call directly
  void removeObjectName (const DString& name, CObjectInfo *info) ; // internal use only

  // Create/Delete ...
  CAttributeInfo *newAttribute ();
  CEnumInfo *newEnum ();
  CUsingInfo *newUsing ();
  CTypedefInfo *newTypedef ();
  CMemberAliasInfo *newMemberAlias (CObjectInfo *);
  void deleteAttribute (const CAttributeInfo *); 
  void deleteEnum (const CEnumInfo *);
  void deleteUsing (const CUsingInfo *);
  void deleteTypedef (const CTypedefInfo *);
  void deleteMemberAlias (const CMemberAliasInfo *);
};

#define INIT_SIZE 20
#define INCR_SIZE 20

inline CStructure::CStructure (CObjectInfo::ObjectId id) :
  CScopeInfo (id),
  _Shared (this),
  _Objects (INIT_SIZE, INCR_SIZE),
  _Attributes (INIT_SIZE, INCR_SIZE),
  _TemplateParams (INIT_SIZE, INCR_SIZE),
  _Functions (INIT_SIZE, INCR_SIZE),
  _Types (INIT_SIZE, INCR_SIZE),
  _Usings (INIT_SIZE, INCR_SIZE),
  _Friends (INIT_SIZE, INCR_SIZE),
  _Namespaces (INIT_SIZE, INCR_SIZE),
  _Aliases (INIT_SIZE, INCR_SIZE)
 {}
 
inline unsigned CStructure::Objects () const
 { return _Objects.length (); }
inline unsigned CStructure::Types () const
 { return _Types.length (); }
inline unsigned CStructure::Attributes () const
 { return _Attributes.length (); }
inline unsigned CStructure::TemplateParams () const
 { return _TemplateParams.length (); }
inline unsigned CStructure::Functions () const
 { return _Functions.length (); }
inline unsigned CStructure::Usings () const
 { return _Usings.length (); }
inline unsigned CStructure::Namespaces () const
 { return _Namespaces.length (); }
inline unsigned CStructure::Friends () const
 { return _Friends.length (); }
inline CObjectInfo *CStructure::Object (unsigned n) const
 { return _Objects.lookup (n); }
inline CObjectInfo *CStructure::Type (unsigned n) const
 { return _Types.lookup (n); }
inline CAttributeInfo *CStructure::Attribute (unsigned n) const
 { return _Attributes.lookup (n); }
inline CTemplateParamInfo *CStructure::TemplateParam (unsigned n) const
 { return _TemplateParams.lookup (n); }
inline CFunctionInfo *CStructure::Function (unsigned n) const
 { return _Functions.lookup (n); }
inline CUsingInfo *CStructure::Using (unsigned n) const
 { return _Usings.lookup (n); }
inline CNamespaceInfo *CStructure::Namespace (unsigned n) const
 { return _Namespaces.lookup (n); }
inline CRecord *CStructure::Friend (unsigned n) const
 { return _Friends.lookup (n); }

inline bool CStructure::isMemberAlias (const CObjectInfo *info) const
 { return (MemberAlias (info)); }
 

} // namespace Puma

#endif /* __CStructure_h__ */
