// This file is part of PUMA.
// Copyright (C) 1999-2003  The PUMA developer team.
//                                                                
// This program is free software;  you can redistribute it and/or 
// modify it under the terms of the GNU General Public License as 
// published by the Free Software Foundation; either version 2 of 
// the License, or (at your option) any later version.            
//                                                                
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the  
// GNU General Public License for more details.                   
//                                                                
// You should have received a copy of the GNU General Public      
// License along with this program; if not, write to the Free     
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
// MA  02111-1307  USA                                            

#ifndef __array_h__
#define __array_h__

#include <assert.h>

namespace Puma {

template <class Item>
class Array {
   protected:
      static const long default_init_size = 5;
      static const long default_increment = 5;

   public:
      Array (long is = default_init_size, long incr = default_increment);
      Array (const Array<Item>& array);
      Array<Item>& operator =(const Array<Item>&);
      ~Array ();
      void append (const Item& item);
      void insert (long index, const Item& item);
      void prepend (const Item& item);
      void remove (long index);
      void reset ();
      Item& get (long index);
      Item& operator[] (long index);
      Item fetch (long index) const;
      Item& lookup (long index) const;
      long length () const;

   private:
      Item* data;
      long size;
      long count;
      long increment;
      
      void check (long wanted);
 };

template <class Item>
inline Array<Item>::Array (long is, long incr)
 {
   count     = 0; 
   size      = is;
   increment = incr;
   data      = new Item[size];
 }

template <class Item>
Array<Item>::Array (const Array<Item>& array)
 {
   count = 0;
   size  = array.size;
   data  = new Item[size];
   increment = array.increment;

   for (int pos = 0; pos < array.length (); pos++)
      append (array.lookup (pos));
 }

template <class Item>
Array<Item>& Array<Item>::operator =(const Array<Item>& array)
 {
   if (data)
      delete[] data;
   count = 0;
   size  = array.size;
   data  = new Item[size];
   increment = array.increment;

   for (int pos = 0; pos < array.length (); pos++)
      append (array.lookup (pos));
   return *this;
 }

template <class Item>
Array<Item>::~Array ()
 {
   if (data)
      delete[] data;
 }

template <class Item>
inline void Array<Item>::check (long wanted)
 {
   if (wanted >= size)
    {
      Item* new_data;

      while (wanted >= size)
      {
         size += increment;
         increment*=2;
      }
      new_data = new Item[size];
      for (int pos = 0; pos < count; pos++)
         new_data[pos] = data[pos];
      delete[] data;
      data = new_data;
    }
 }
   
template <class Item>
inline void Array<Item>::append (const Item& item)
 {
   check (count);
   data[count++] = item;
 }

template <class Item>
void Array<Item>::prepend (const Item& item)
 {
   insert (0, item);
 }

template <class Item>
void Array<Item>::insert (long index, const Item& item)
 {
   check (count);
   for (int pos = count; pos > index; pos--)
      data[pos] = data[pos - 1];
   data[index] = item;
   count++;
 }

template <class Item>
inline Item& Array<Item>::get (long index)
 {
   check (index);
   if (index >= count)
      count = index + 1;
   return data[index];
 }

template <class Item>
inline Item& Array<Item>::operator[] (long index)
 {
   return get (index);
 }

template <class Item>
inline Item Array<Item>::fetch (long index) const
 {
   assert(index < count);
   return data[index];
 }

template <class Item>
inline long Array<Item>::length () const
 {
   return count;
 }

template <class Item>
inline void Array<Item>::remove (long index)
 {
   if (index < count && count > 0)
    {
      for (int pos = index; pos < count - 1; pos++)
         data[pos] = data[pos + 1];
      count--;
    }
 }

#ifndef __puma
template <>
inline void Array<int>::reset () {
   count = 0; 
}
#endif

template <class Item>
void Array<Item>::reset ()
 {
   if (data)
      delete[] data;
   count     = 0; 
   size      = default_init_size;
   increment = default_increment;
   data      = new Item[size];
 }

template <class Item>
inline Item& Array<Item>::lookup (long index) const
 {
   assert(index >= 0 && index < count);
//   if (index >= count) index = count - 1; 
//   if (index < 0)      index = 0;
   return data[index];
 }


} // namespace Puma

// Array specialization for pointer types sharing method definitions

#ifndef __puma

#include "Puma/VoidPtrArray.h"

namespace Puma {


template <class Item>
class Array<Item*> : private VoidPtrArray {
  typedef VoidPtrArray Base;

public:
  Array (long is = default_init_size, long incr = default_increment) :
    Base (is, incr) {}
  Array (const Array<Item*> &array) : 
    Base (array) {}
  Array<Item*> &operator= (const Array<Item*> &array) 
   { return (Array<Item*>&)Base::operator= (array); }

  ~Array () {}

  void append (const Item *item)           { Base::append ((void*)item); }
  void insert (long idx, const Item *item) { Base::insert (idx, (void*)item); }
  void prepend (const Item *item)          { Base::prepend ((void*)item); }
  void remove (long idx)                   { Base::remove (idx); }
  void reset ()                            { Base::reset (); }

  Item *&get (long idx)          { return (Item*&)Base::get (idx); }
  Item *&operator[] (long idx)   { return (Item*&)Base::operator[] (idx); }
  Item *fetch (long idx) const   { return (Item*)Base::fetch (idx); }
  Item *&lookup (long idx) const { return (Item*&)Base::lookup (idx); }
  long length () const           { return Base::length (); }
};


} // namespace Puma

#endif /* __puma */

#endif /* __array_h__ */
